#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <curl/curl.h>
#include <jsoncpp/json/json.h>

#define CURL_OK 0

// Defines non-standard Quil gates used in code generation
static std::string preamble = "DEFCIRCUIT I2 q0 q1:\n\tI q0; I q1;\n\n\nDEFCIRCUIT I3 q0 q1 q2:\n\tI q0; I q1; I q2;\n\n\nDEFCIRCUIT QMOVE q0 c0:\n\tRESET q0;\n\tJUMP-UNLESS @end-reset c0;\n\tX q0;\n\tLABEL @end-reset;\n\n";

static size_t qvm_request_callback(char* buffer, size_t buffer_size,
                                   size_t element_size, void* userdata) {
    FILE *json_data = (FILE*)userdata;
    int bytes_read = fread(buffer, buffer_size, element_size, json_data);
    return bytes_read;
}

static size_t qvm_response_callback(char* results, size_t result_len,
                                    size_t element_size, void* userdata) {
    std::string * result_store = (std::string *)userdata;
    *(result_store) = std::string(results, result_len*element_size);
    return result_store->length();
}

extern "C" int quant_com(const char * program_name, const char * quil_program) {

        std::string programName = std::string(program_name);
        std::string quilProgram = std::string(quil_program);
        std::string program_json = "{\"type\": \"multishot\", \"trials\": 1, \"addresses\": {\"" + programName + "_ro\": true}, \"quil-instructions\":\"" + preamble + quilProgram + "\n" + programName + "\"}";
        // std::cout << program_json << "\n";
        // Initalize cURL to send message to quantum simulator
        CURL * curl_session = curl_easy_init();

        if(curl_session) {
            // Build request JSON and setup callbacks to send/receive data
            FILE * json_buffer = fmemopen((char *)program_json.c_str(),
                                          program_json.length(), "r");
            std::string result_json;
            curl_easy_setopt(curl_session, CURLOPT_READFUNCTION, qvm_request_callback);
            curl_easy_setopt(curl_session, CURLOPT_READDATA, (void *)json_buffer);
            curl_easy_setopt(curl_session, CURLOPT_WRITEFUNCTION, qvm_response_callback);
            curl_easy_setopt(curl_session, CURLOPT_WRITEDATA, (void *)&result_json);

            // Build HTTP request for quantum simulator
            struct curl_slist *header = NULL;
            header = curl_slist_append(header, "Content-Type: application/json");
            curl_easy_setopt(curl_session, CURLOPT_HTTPHEADER, header);
            curl_easy_setopt(curl_session, CURLOPT_POST, 1L);
            curl_easy_setopt(curl_session, CURLOPT_URL, "http://localhost:5000");

            // Send message to quantum simulator

            CURLcode result = curl_easy_perform(curl_session);
            if(result == CURL_OK) {
                // Convert JSON string to object
                Json::Value result_object;
                std::stringstream json_stream(result_json);
                json_stream >> result_object;
                // std::cout << result_object << "\n";

                // Check if the returning JSON was an error message or not
                std::string error_message;
                if((error_message = result_object.get("error_type", "").asString())
                    != "") {
                    std::cerr << "Quantum program failed: "
                              << error_message
                              << "\n";
                    curl_easy_cleanup(curl_session);
                    exit(-1);
                }
                else {
                    Json::Value result_bits = result_object[programName + "_ro"][0];
                    int result_value = 0;
                    // std::cout << std::to_string(result_bits.size());
                    for(int i = result_bits.size()-1;
                            i >= 0;
                            i--) {
                                /*
                        std::cout << "result_bits["+ std::to_string(i)+"]= "
                                     + result_bits[i].asString() + "\n";
                                     */
                        result_value = (result_value << 1) ^
                                        result_bits[i].asInt();
                    }
                    /*
                    std::cout << "result_value = " +
                        std::to_string(result_value) + "\n";
                        */
                    curl_easy_cleanup(curl_session);
                    return result_value;
                }
            } else {
                std::cerr << "Failed to get result from quantum server\n";
                curl_easy_cleanup(curl_session);
                exit(-1);
            }
        }

        curl_easy_cleanup(curl_session);
        exit(-1);
    }
